/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.taglet;

import java.lang.reflect.Field;

import com.sun.javadoc.Tag;
import com.sun.javadoc.Doc;
import com.sun.javadoc.MethodDoc;
import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.PackageDoc;
import com.sun.javadoc.ProgramElementDoc;

/**
 * Doclet を処理するプログラムで共通して使用される簡易メソッド群(ユーティリティクラス)です。
 *
 * @version  4.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public final class DocletUtil {
	/** リターンコード  System.getProperty("line.separator")  */
	public static final String CR = System.getProperty("line.separator");

	/**
	 * すべてが staticメソッドなので、コンストラクタを呼び出さなくしておきます。
	 *
	 */
	private DocletUtil() {}

	/**
	 * 入力文字列の HTML 文字をフィルタリングします。
	 * "&" は、"&amp;#38;" 、"<" は、"&amp;lt;" 、">" は、"&amp;gt;" に変換します。
	 * StringUtil.htmlFilter との違いは、';' → "&#59;" への変換があることです。
	 *
	 * @og.rev 5.5.4.1 (2012/07/06) DocletUtil.htmlFilter → StringUtil.htmlFilter に変更のため、廃止
	 *
	 * @param input String
	 *
	 * @return 変換後文字列
	 */
//	public static String htmlFilter( final String input ) {
//		if( input == null ) { return ""; }
//
//		int len = input.length();
//
//		StringBuilder rtn = new StringBuilder( len + 50 );
//		char ch;
//		for(int i=0; i<len; i++) {
//			ch = input.charAt(i);
//			switch( ch ) {
//				case '<'  : rtn.append("&lt;");   break;
//				case '>'  : rtn.append("&gt;");   break;
//				case '"'  : rtn.append("&quot;"); break;
//				case '\'' : rtn.append("&#39;");  break;
//				case '&'  : rtn.append("&amp;");  break;
//				case ';'  : rtn.append("&#59;");  break;
//				default   : rtn.append(ch);
//			}
//		}
//		return( rtn.toString() );
//	}

	/**
	 * target 文字列に含まれる from 文字列を to 文字列に置き換えます。
	 *
	 * @param	target	元の文字列
	 * @param	from	置換元FROM
	 * @param	to		置換先TO
	 *
	 * @return	変換後文字列
	 */
	public static String replace( final String target,final String from,final String to ) {
		if( target == null || from == null || to == null ) { return target; }
		StringBuilder strBuf = new StringBuilder( 200 );

		int start = 0;
		int end   = target.indexOf( from,start );
		while( end >= 0  ) {
			strBuf.append( target.substring( start,end ) );
			strBuf.append( to );
			start = end + from.length();
			end   = target.indexOf( from,start );
		}
		strBuf.append( target.substring( start ) );

		return strBuf.toString();
	}

	/**
	 * コメント部の文字列を取得します。
	 *
	 * @og.rev 5.5.4.1 (2012/07/06) コメントは文字列でなく、Tag配列として処理させる。
	 *
	 * @param cmnt String
	 *
	 * @return コメント文字列
	 */
//	public static String commentText( final String cmnt ) {
//		if( cmnt == null ) { return ""; }
//
//		String rtn = cmnt;
//		int indx = cmnt.indexOf( "{@value}" );
//		if( indx >= 0 ) {
//			rtn = cmnt.substring( indx+8 );	// {@value} 以前を削除
//		}
//		return htmlFilter( rtn );
//	}

	/**
	 * セッターメソッドの setXXXX の set を削除し、次の文字を小文字化します。
	 * つまり、セッターメソッドから属性値を推測します。
	 * (超特殊処理)セッターメソッドのset以下２文字目が大文字の場合は、
	 * １文字目も大文字と考えて小文字化を行いません。
	 * 例えば、setSYS や setUSER など、RequestValueTag.javaに使用するケースです。
	 *
	 * @param target 処理対象となる文字列
	 *
	 * @return オプション文字列
	 */
	public static String removeSetter( final String target ) {
		if( target != null && target.startsWith( "set" ) ) {
			char[] chs = target.toCharArray();
			if( chs.length > 4 && !( chs[4] >= 'A' && chs[4] <= 'Z' ) ) {
				chs[3] = Character.toLowerCase( chs[3] ) ;
			}
			return new String( chs,3,chs.length-3 );
		}
		return target;
	}

	/**
	 * オプション配列文字列より、指定のキーに対応するオプション値を返します。
	 *
	 * @param	key		キー
	 * @param	options	オプション配列文字列
	 *
	 * @return	オプション文字列
	 */
	public static String getOption( final String key , final String[][] options ) {
		String rtn = "";
		if( key == null || options == null ) { return rtn; }

		for( int i=0; i<options.length; i++ ) {
			if( key.equalsIgnoreCase( options[i][0] ) ) {
				rtn = options[i][1];
				break ;
			}
		}
		return rtn ;
	}

	/**
	 * {&#064;og.value package.class#field} 形式のvalueタグを文字列に置き換えます。
	 *
	 * 処理的には、リフレクションで、値を取得します。値は、staticフィールドのみ取得可能です。
	 *
	 * @og.rev 5.5.4.1 (2012/07/06) 新規追加
	 * @og.rev 5.5.5.6 (2012/08/31) クラス名の取得で、ProgramElementDoc で処理するように変更
	 *
	 * @param tag Tagオブジェクト
	 *
	 * @return valueタグの解析結果の文字列
	 */
	public static String valueTag( final Tag tag ) {
		String txt = tag.text();
		if( txt != null ) {
			String cls = null;		// package.class
			String fld = null;		// field

			// package.class#field 形式の解析。
			int adrs = txt.indexOf('#') ;
			if( adrs > 0 ) {
				cls = txt.substring( 0,adrs );		// package.class
				fld = txt.substring( adrs+1 );		// field
			}
			else if( adrs == 0 ) {
				fld = txt.substring( 1 );			// #field
			}
			else {
				String errMsg = "警告:{@value package.class#field} 形式の フィールド名 #field がありません。" + CR
								+ tag.position() + " : " + txt + CR ;
				System.err.println( errMsg );
				// # を付け忘れたと考え、自分自身のクラスを利用
				fld = txt;							// field
			}

			// package.class をきちんと作成する。
			Doc doc = tag.holder();
//			if( doc.isMethod() ) {
//				MethodDoc mdoc = (MethodDoc)doc;
//				ClassDoc cdoc = mdoc.containingClass();
//				if( cls == null ) {					// package.class が登録されていない場合。
//					cls = cdoc.qualifiedName() ;
//				}
//				else if( cls.indexOf('.') < 0 ) {	// class のみが登録されている場合。findClass で、package 込の正式クラス名を検索する。
//					ClassDoc pdoc = cdoc.findClass( cls );
//					cls = pdoc.qualifiedName() ;
//				}
//			}

			// 5.5.5.6 (2012/08/31) ProgramElementDoc で処理するように変更
			if( doc instanceof ProgramElementDoc ) {
				ProgramElementDoc pdoc = (ProgramElementDoc)doc;
				ClassDoc cdoc = pdoc.containingClass();
				if( cdoc != null ) {
					if( cls == null ) {					// package.class が登録されていない場合。
						cls = cdoc.qualifiedName() ;
					}
					else if( cls.indexOf('.') < 0 ) {	// class のみが登録されている場合。findClass で、package 込の正式クラス名を検索する。
						ClassDoc fdoc = cdoc.findClass( cls );
						if( fdoc != null ) {
							cls = fdoc.qualifiedName() ;
						}
					}
				}
				else {
					if( cls == null ) {					// package.class が登録されていない場合。
						cls = pdoc.qualifiedName() ;
					}
				}
			}

			try {
				Field fldObj = Class.forName( cls ).getDeclaredField( fld );
				// privateフィールドへのアクセス。(セキュリティーマネージャーによってアクセス制限がかけられていない場合)
				if( !fldObj.isAccessible() ) { fldObj.setAccessible( true ); }
				txt = String.valueOf( fldObj.get( null ) );		// static フィールドは、引数 null で値を取得
			}
			catch( Exception ex ) {
				String errMsg = "警告:{@value package.class#field} 形式の対象がありません。" + CR
								+ tag.position() + " : " + tag.text() + CR
								+ ex.getMessage() + CR;
				System.err.println( errMsg );
			}
		}
		return txt ;
	}
}
